#!/usr/bin/python3
__version__ = "1.0"

import coralme.builder.main
import coralme.builder.dictionaries
import coralme.builder.organism
import coralme.builder.homology
import coralme.builder.curation
import coralme.builder.helper_functions
import coralme.builder.gpr
import coralme.builder.troubleshooting
import coralme.builder.notes

# 1st step
import coralme.builder.flat_files
import coralme.builder.preprocess_inputs
import coralme.builder.dna_replication

# 2nd step
import coralme.builder.ribosome
import coralme.builder.trna_charging
import coralme.builder.transcription
import coralme.builder.modifications
import coralme.builder.translation
import coralme.builder.translocation
import coralme.builder.formulas
import coralme.builder.compartments
